@testable import App
import FluentPostgreSQL
import Vapor
import XCTest

// swiftlint:disable force_try force_unwrapping
final class TodoTests: XCTestCase {

    let todosURI = "/todos/"
    let todoTitle = "My Todo!"
    var app: Application!
    var conn: PostgreSQLConnection!

    override func setUp() {
        try! Application.reset()
        app = try! Application.testable()
        conn = try! app.newConnection(to: .psql).wait()
    }

    override func tearDown() {
        conn.close()
    }

    func testTodosCanBeRetrievedFromAPI() throws {
        let todo = try Todo.create(title: todoTitle, on: conn)
        _ = try Todo.create(on: conn)

        let todos = try app.getResponse(to: todosURI, decodeTo: [Todo].self)

        XCTAssertEqual(todos.count, 2)
        XCTAssertEqual(todos[0].title, todoTitle)
        XCTAssertEqual(todos[0].id, todo.id)
    }

    func testTodoCanBeSavedWithAPI() throws {
        let todo = Todo(title: todoTitle)
        let receivedTodo = try app.getResponse(to: todosURI, method: .POST, headers: ["Content-Type": "application/json"], data: todo, decodeTo: Todo.self)

        XCTAssertEqual(receivedTodo.title, todoTitle)
        XCTAssertNotNil(receivedTodo.id)

        let todos = try app.getResponse(to: todosURI, decodeTo: [Todo].self)

        XCTAssertEqual(todos.count, 1)
        XCTAssertEqual(todos[0].title, todoTitle)
        XCTAssertEqual(todos[0].id, receivedTodo.id)
    }

    func testTodoCanBeDeletedWithAPI() throws {
        let todo = try Todo.create(title: todoTitle, on: conn)

        let response = try app.sendRequest(to: "\(todosURI)\(todo.id!)", method: .DELETE)

        XCTAssertEqual(response.http.status, .noContent)

        let todos = try app.getResponse(to: todosURI, decodeTo: [Todo].self)

        XCTAssertEqual(todos.count, 0)
    }

    func testGettingASingleTodoFromTheAPI() throws {
        let todo = try Todo.create(title: todoTitle, on: conn)
        let receivedTodo = try app.getResponse(to: "\(todosURI)\(todo.id!)", decodeTo: Todo.self)

        XCTAssertEqual(receivedTodo.title, todoTitle)
        XCTAssertEqual(receivedTodo.id, todo.id)
    }

    func testLinuxTestSuiteIncludesAllTests() throws {
        #if os(macOS) || os(iOS) || os(tvOS) || os(watchOS)
        let thisClass = type(of: self)
        let linuxCount = thisClass.allTests.count
        let darwinCount = Int(thisClass.defaultTestSuite.testCaseCount)

        XCTAssertEqual(linuxCount, darwinCount, "\(darwinCount - linuxCount) tests are missing from allTests")
        #endif
    }

    static let allTests = [
        ("testLinuxTestSuiteIncludesAllTests", testLinuxTestSuiteIncludesAllTests),
        ("testTodosCanBeRetrievedFromAPI", testTodosCanBeRetrievedFromAPI),
        ("testTodoCanBeSavedWithAPI", testTodoCanBeSavedWithAPI),
        ("testTodoCanBeDeletedWithAPI", testTodoCanBeDeletedWithAPI),
        ("testGettingASingleTodoFromTheAPI", testGettingASingleTodoFromTheAPI)
    ]
}
